/*

PlayerState contains various global state of the playing system, such as volume,
current subtune, a reference to the playlist pane etc. It is in effect a singleton
but is not accessible to other systems except when it is fed to them through a
function.

A ListPane is a list of files that can be manipulated and rendered.

Equaliser is also a graphic object that gets fed audio data, and when rendered
presents that data as equalizer bars

The PluginSystem knows how to handle files.

SongInfoPane displays songinfo


*/


#include <SDL.h>
//#include "plugin.h"
#include <vector>
#include "FileList.h"
#include "ZipFileList.h"
#include "SDLScreen.h"

#include "MusicPlayer.h"
#include "Input.h"
#include "Fifo.h"
#include "Equalizer.h"

#ifndef _WIN32
#define stricmp strcasecmp
#define strnicmp strncasecmp
#include <unistd.h>
#else
#include <direct.h>
#endif

using namespace std;

MusicPlayer *mp = NULL;

static int is_ext(const char *s, const char *ext)
{
	const char *sext = strrchr(s, '.');
	if(!sext)
		sext = "";
	return (stricmp(ext, sext) == 0);
}


int plugin_filter_cb(const char *name)
{
	//printf("Checking %s\n", name);

	if(mp->CanHandle(name))
		return 1;

	if(is_ext(name, ".zip"))
		return 1;

	return 0;
}



ListPane *pane;
SongInfoPane *songInfo;
LedControl *lcd;
Equalizer *eq = NULL;
ScreenSettings ss;
#define ABS(x) ((x) < 0 ? (-x) : (x))

void render_eq()
{
	short buff[4096];
	short buff2[2048];

	int l = mp->GetNextSamples(buff, sizeof(buff)) / 2;

	int avg0 = 0;
	int avg1 = 0;

	if(l)
	{
		int i;
		for(i=0; i<l; i+=2)
		{
			avg0 += ABS(buff[i]);
			avg1 += ABS(buff[i+1]);
		}
		//avg0 /= l;
		//avg1 /= l;
		//cprintf("%d %d\n", l, avg0*2/l);
		lcd->SetLed(LedControl::BAR0, avg0 / (l*256));
		lcd->SetLed(LedControl::BAR1, avg1 / (l*256));

		for(i=0; i<l/4; i++)
			buff2[i] = (buff[i*2] + buff[i*2+1]) / 2;

		if(eq)
			eq->WriteSamples(buff2, l/4);


	}



}


void layout(ScreenSettings &ss)
{
}

static int bcount = 0;
extern "C" {

void busy(int x)
{
	if(x)
		bcount = (bcount+1) % 6;
	else
		bcount = 0;
	lcd->SetLed(LedControl::BUSY, bcount);
	lcd->Render();
}


}

int main(int argc, char **argv)
{
	ss.bevellight = 0xB2DBF8;
	ss.bevelmid = 0x5084A4;
	ss.beveldark = 0x1F596C;
	ss.bgcol = 0x000000;
	ss.textcol = 0xFFFFFF;

	int hmargin = 4;
	int vmargin = 4;
	int sep = 8;



#ifdef _WIN32

	ss.font0 = "Fixed.font";
	ss.font1 = "Proportional.font";
	ss.width = 640;
	ss.height = 570;
	char *led = "bigled.bmp";

#else

	ss.font0 = "";
	ss.font1 = "";
	ss.width = 320;
	ss.height = 240;
	char *led = "led.bmp";

#endif

	g_screen.SetSettings(&ss);

	printf("Starting\n");
	g_screen.Init();
	printf("OK\n");

	pane = new ListPane(0, 0, 0, 0);
	songInfo = new SongInfoPane(0, 0, 0, 0);
	lcd = new LedControl(led, 0, 0);
	printf("OK\n");

	int fh = g_screen.GetFontHeight();

	int w = ss.width - hmargin*2;
	int h = ss.height - vmargin|2;

	lcd->Move(w - lcd->Width() + hmargin, fh+sep);
	pane->Move(hmargin, fh+sep);
	pane->Resize(w - lcd->Width() - sep, h - (fh+sep));
	pane->Clear(true);
	songInfo->Move(w - lcd->Width() + hmargin, lcd->Height() + (fh+sep) + sep);
	songInfo->Resize(lcd->Width(), fh * 9);
	songInfo->Clear(true);
	songInfo->Render();


	int y = lcd->Height() + (fh+sep) + sep + sep + fh * 9;

	eq = new Equalizer(w - lcd->Width() + hmargin, y, lcd->Width(), lcd->Width()/3);//lcd->Width(), h-y);
	eq->Clear(true);
	eq->Render();

	g_screen.Update();

	//g_screen.CreateConsole(w - lcd->Width() + hmargin, y, lcd->Width(), h-y); 

	printf("OK %d\n", songInfo->Height());

	//ListPane *pane = new ListPane(4,16,ss.width - 120, ss.height - 20);
	//TextDisplay *text = new TextDisplay(ss.width-104, 26, 104, 56);
	//LedControl *lcd = new LedControl("led5.bmp", 212, 90);

	mp = new MusicPlayer();
	Input *input = new Input();


	lcd->Render();
	//text->PutText(1,1,"B = Enter/Play");
	//text->PutText(1,2,"X = Parent dir");
	//text->PutText(1,3,"Y/A = Page Up/Down");
	//text->PutText(1,4,"Stick L/R = Subtune");
	//text->PutText(1,5,"Start = Quit");
	//text->Render();
	//songInfo->Render();

	char curdir[256];
	getcwd(curdir, sizeof(curdir));
	pane->SetList(new FileList(curdir, &plugin_filter_cb));
	pane->Render();

	int selected = -1;
	int repeat = 0;
	int shuffle = 0;
	int parent_selected = 0;
	IFileList *parent_file_list = NULL;
	int last_sec = -1;

	int paused = 0;
	int playing = 0;
	int oldplaying = -1;

	lcd->SetLed(LedControl::SONGNO, 1);
	lcd->SetLed(LedControl::MAXSONG, 1);
	lcd->SetLed(LedControl::VOLUME, mp->GetVolume());

	int ucount = 0;

	while(1)
	{
		IFileList *file_list = pane->GetList();

		while(selected == -1)
		{
			SDL_Delay(50);
			pane->Render();
			songInfo->Render();
			lcd->Render();
			//songInfo->SetFormat(mp->GetFormat());
			//printf("%s\n", mp->GetFormat());

#ifndef _WIN32
			if(!mp->GetHeavy())
#endif
			{
				if(eq)
					eq->Render();
				render_eq();
			}
			//if(ucount++ == 2)
			//{
				
			//	ucount = 0;
			//}

			if(paused)
			{
				if(((paused-1) % 15) == 0)
					lcd->SetLed(LedControl::PLAYING, (paused-1) % 30 );
				paused++;
			}

			if(playing != oldplaying)
			{
				lcd->SetLed(LedControl::PLAYING, playing );	
				oldplaying = playing;
			}



			int sec = mp->GetCSeconds() / 10;

			if(last_sec != sec)
			{
				lcd->SetLed(LedControl::LENGTH, sec);
				last_sec = sec;
			}

			int key = input->GetKey(40);
			if(key)
			{
				switch(key)
				{
				case Input::PGUP:
					pane->DoCommand(ListPane::PGUP);
					break;
				case Input::PGDOWN:
					pane->DoCommand(ListPane::PGDN);
					break;
				case Input::UP:
					pane->DoCommand(ListPane::UP);
					break;
				case Input::DOWN:
					pane->DoCommand(ListPane::DOWN);
					break;
				case Input::FF:
					mp->SetTune(mp->GetTune() + 1);
					lcd->SetLed(LedControl::SONGNO, mp->GetTune()+1);
					break;
				case Input::REW:
					mp->SetTune(mp->GetTune() - 1);
					lcd->SetLed(LedControl::SONGNO, mp->GetTune()+1);
					break;
				case Input::ENTER:
					selected = pane->GetMarked();;
					break;
				case Input::BACK:
					selected = -2;
					break;
				case Input::VOLUP:
					mp->SetVolume(mp->GetVolume() + 1);
					lcd->SetLed(LedControl::VOLUME, mp->GetVolume());
					break;
				case Input::VOLDOWN:
					mp->SetVolume(mp->GetVolume() - 1);
					lcd->SetLed(LedControl::VOLUME, mp->GetVolume());
					break;
				case Input::QUIT:
					SDL_CloseAudio();
					SDL_Quit();
					return 0;
				case Input::SHUFFLE:
					shuffle ^= 1;
					lcd->SetLed(LedControl::SHUFFLE, shuffle);
					break;
				case Input::REPEAT:
					repeat ^= 1;
					lcd->SetLed(LedControl::REPEAT, repeat);
					break;
				case Input::PAUSE:
					paused = mp->Pause();
					break;
				case Input::BLENDUP:
#ifdef GP2X
					if(mp->GetBlend())
						mp->SetBlend(0);
					else
						mp->SetBlend(50);
#else
					mp->SetBlend(mp->GetBlend() + 25);
#endif
					lcd->SetLed(LedControl::EQ0, mp->GetBlend() / 50);
					lcd->SetLed(LedControl::EQ1, 5 - (mp->GetBlend() / 50));

					break;
				case Input::BLENDDOWN:
					mp->SetBlend(mp->GetBlend() - 25);
					lcd->SetLed(LedControl::EQ0, mp->GetBlend() / 50);
					lcd->SetLed(LedControl::EQ1, 5 - (mp->GetBlend() / 50));
					break;
				}
			}
		}

		if(selected < 0)
		{
			int rc = file_list->Enter(-1);
			if(rc < 0)
			{
				delete file_list;
				file_list = parent_file_list;
				rc = parent_selected;
				parent_selected = 0;
				parent_file_list = NULL;
			}
			pane->SetList(file_list);
			if(rc >= 0)
				pane->SetMarked(rc);
		}
		else
		{
			vector<FileRef> &finfo = file_list->GetNames();
			const char *fname = file_list->GetFileName(selected);

			int fsize = finfo[selected].size;
			const char *name = finfo[selected].name;

			const char *ext = strrchr(name, '.');
			if(!ext)
				ext = "";

			if(fsize == -1)
			{
				file_list->Enter(selected);
				pane->SetList(file_list);
				selected = -1;
			}
			else
			{
				if(stricmp(ext, ".zip") == 0)
				{
					parent_file_list = file_list;
					parent_selected = selected;
					file_list = new ZipFileList((char *)fname);
					pane->SetList(file_list);
					selected = -1;
				}
				else
				{
					bool ok = false;
					if(fname)
						ok = mp->PlaySong(fname);
					if(!ok)
					{
						char *tmp = (char *)malloc(fsize);
						file_list->Read(selected, tmp, 0, fsize);
						ok = mp->PlaySong(name, tmp, fsize);
						
						free(tmp);
					}
					playing = (int)ok;
					if(ok)
					{
						lcd->SetLed(LedControl::MAXSONG, mp->GetMaxTunes());
						lcd->SetLed(LedControl::SONGNO, mp->GetTune()+1);
						songInfo->SetName(mp->GetName());
						songInfo->SetArtist(mp->GetArtist());
						songInfo->SetCopyright(mp->GetCopyright());
						songInfo->SetLength(mp->GetLength());
						songInfo->SetFormat(mp->GetFormat());
					}
					else
					{
						cprintf("Could not play file\n");
					}
				}
			}

		}
		selected = -1;
	}

	return 0;
}
