/*  Copyright 2006 Jonas Minnberg

    This file is part of OldPlay - a portable, multiformat musicplayer.

    OldPlay is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    OldPlay is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OldPlay; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include <SDL.h>
#include <vector>
#include <sys/stat.h>

#ifdef USE_DEV_MIXER
#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/soundcard.h>
#endif

#include "MusicPlayer.h"
#include "FileList.h"
#include "Fifo.h"
#include "util.h"
#include "plugin.h"

int cprintf(char *fmt, ...);

#ifdef _WIN32
#include <windows.h>
void *dlopen(const char *filename, int flag)
{
	return (void *)LoadLibraryA(filename);
}

void *dlsym(void *handle, const char *symbol)
{
	HINSTANCE h = (HINSTANCE)handle;
	return (void *)GetProcAddress(h, symbol);
}
#else
#include <dlfcn.h>
#define stricmp strcasecmp
#define strnicmp strncasecmp
#endif

#ifdef GP2X
#define BLOCK_SIZE 2048
#define BLOCK_COUNT 8
#else
#define BLOCK_SIZE 8192
#define BLOCK_COUNT 4
#endif

struct sound_plugin *current_plugin = NULL;
static short audiobuf[BLOCK_SIZE * BLOCK_COUNT];

void MusicPlayer::fill_audio(BlockFifo *af, int max)
{
	if(current_plugin && current_plugin->fill_buffer)
	{
		int b = af->LeftToWrite();

		if(b)
		{
			if(max && b > max) b = max;
            // *note* this asks for an amount of bytes, not samples (UGH!)
			int rc = current_plugin->fill_buffer(audiobuf, BLOCK_SIZE * b);
			if(!rc)
			{
				if(current_plugin && current_plugin->close)
					current_plugin->close();
				current_plugin = NULL;
				af->Reset();
				fprintf(stderr, "MusicPlayer::fill_audio - DONE\n");
			}
			else
				af->Write(audiobuf, rc / BLOCK_SIZE);

		}
	}

}

static char go_on = 1;

int MusicPlayer::GetNextSamples(short *dest, int len) //Unimportant
{
	int blocks = (len+BLOCK_SIZE-1)/BLOCK_SIZE;

	if(blocks > audioFifo->LeftToRead())
		blocks = audioFifo->LeftToRead();

	if(current_plugin && current_plugin->freq < 30000)
		audioFifo->Read(dest, blocks, 1);
	else
		audioFifo->Read(dest, blocks+1, 2);
	return blocks * BLOCK_SIZE;
}

void MusicPlayer::audio_cb(void *userdata, Uint8 *stream, int len)
{
	MusicPlayer *mp = (MusicPlayer *)userdata;
	if(mp->paused)
	{
		memset(stream, 0, len);
		return;
	}

	//fprintf(stderr, "%d bytes, %d blocks available\n", len, ps.audioFifo->LeftToRead());
	if(mp->audioFifo->LeftToRead())
	{
        int xxx;
#ifdef USE_DEV_MIXER
		xxx = mp->audioFifo->Read(stream, 1);
#else
		mp->audioFifo->Read16(stream, 1, mp->vol * 0x10000 / 10);
#endif
        /*
        FILE *fp = fopen("audio_cb_out","ab");
        fwrite(stream,1,xxx*BLOCK_SIZE,fp);
        fclose(fp);
        */

        if(mp->blend)
		{
			int b = mp->blend;
			int l,r,i;
			short *s = (short *)stream;
			for(i=0; i<len>>2; i++, s += 2)
			{
				r = (s[0] * b + s[1] * (256-b)) >> 8;
				l = (s[1] * b + s[0] * (256-b)) >> 8;
				s[0] = l;
				s[1] = r;
			}
		}

		mp->total_samples += len;
        if (current_plugin->length > 0)
            mp->songend_samples = (current_plugin->length * 4) / 1000 * current_plugin->freq;
		if(mp->songend_samples < 0 || (mp->songend && mp->total_samples > mp->songend_samples))
		{
            if (current_plugin->subtunes > 1+current_plugin->tune)
            {
                printf("MusicPlayer::audio_cb - Going to next subtune %d to %d!\n",current_plugin->tune,current_plugin->tune+1);
                mp->SetTune(current_plugin->tune+1);
            }
            else
            {
                printf("MusicPlayer::audio_cb: Closing song\n");
                if(current_plugin && current_plugin->close)
                    current_plugin->close();
                mp->audioFifo->Reset();
                current_plugin = NULL;
            }
		}
	}
	else
		memset(stream, 0, len);

	if(!go_on)
		return;
	fill_audio(mp->audioFifo, 2);
	if(!current_plugin)
		mp->playing = false;
}

int mixer_fd = -1;
bool aopen = false;
int current_freq;

int MusicPlayer::InitSound(int freq)
{
#ifdef USE_DEV_MIXER
	if(mixer_fd < 0)
		mixer_fd = open("/dev/mixer", O_RDWR);
	int mvol =(((vol*0x50)/10)<<8)|((vol*0x50)/10);   
	ioctl(mixer_fd, SOUND_MIXER_WRITE_SPEAKER, &mvol); 
	ioctl(mixer_fd, SOUND_MIXER_WRITE_VOLUME,  &mvol); 
#endif

	SDL_AudioSpec desired, obtained;
	memset(&desired, 0, sizeof(desired));
	desired.format = SDL_BYTEORDER == SDL_BIG_ENDIAN ? AUDIO_S16MSB : AUDIO_S16LSB;
	desired.freq = freq;
	desired.channels = 2;
	desired.samples = BLOCK_SIZE/4;
	desired.callback = audio_cb; 
	desired.userdata = this;

	fprintf(stderr, "MusicPlayer::InitSound - Open Audio\n");
	if(aopen)
		SDL_CloseAudio();
	else
		audioFifo = new BlockFifo(BLOCK_SIZE, BLOCK_COUNT);

	audioFifo->Reset();
	aopen = true;
	int tries = 5;

	while(SDL_OpenAudio(&desired, &obtained) < 0)
	{
      	//cprintf("Couldn't open SDL audio: %s\n", SDL_GetError());
      	fprintf(stderr, "MusicPlayer::InitSound - Couldn't open SDL audio: %s\n", SDL_GetError());
		SDL_Delay(500);

		if(tries == 0)
			exit(-1);
		tries--;
	}

	SDL_PauseAudio(0);
	fprintf(stderr, "MusicPlayer::InitSound - Audio Opened\n");

	return obtained.freq;
}

#ifdef STATIC
extern "C" {
sound_plugin *sidplay_init_sound_plugin();
sound_plugin *modplay_init_sound_plugin();
sound_plugin *uade_init_sound_plugin();
sound_plugin *gme_init_sound_plugin();
sound_plugin *mad_init_sound_plugin();
sound_plugin *pokecubed_init_sound_plugin();
}
#endif

int MusicPlayer::CanHandle(const char *name)
{
	for(unsigned int i=0; i<sound_plugins.size(); i++)
		if(sound_plugins[i]->can_handle(name))
			return 1;
	return 0;
}

MusicPlayer::MusicPlayer()
{
#ifdef STATIC
	struct sound_plugin *plugin;
	if(plugin = sidplay_init_sound_plugin())
		sound_plugins.push_back(plugin);
	if(plugin = gme_init_sound_plugin())
		sound_plugins.push_back(plugin);
	if(plugin = modplay_init_sound_plugin())
		sound_plugins.push_back(plugin);
	if(plugin = mad_init_sound_plugin())
		sound_plugins.push_back(plugin);
    if(plugin = pokecubed_init_sound_plugin());
        sound_plugins.push_back(plugin);
#ifndef _MSC_VER
	if(plugin = uade_init_sound_plugin())
		sound_plugins.push_back(plugin);
#endif

#else
	FileList *flist = new FileList("plugins");

	cprintf("%p %d\n", flist, flist ? flist->Size() : -1);
	for(int i=0; i<flist->Size(); i++)
	{
		const char *fn = flist->GetFileName(i);
		void *ptr = dlopen(fn, 0);

		fprintf(stderr, "MusicPlayer::MusicPlayer - Plugin %s:", fn);
		if(ptr)
		{
			struct sound_plugin *(*init)() = (struct sound_plugin *(*)())dlsym(ptr, "init_sound_plugin");
			if(init)
			{
				struct sound_plugin *plugin = init();
				sound_plugins.push_back(plugin);
				fprintf(stderr, "MusicPlayer::MusicPlayer - Loaded\n");
			}
			else
				fprintf(stderr, "MusicPlayer::MusicPlayer - Init failed!\n");
		}
		else
			fprintf(stderr, "MusicPlayer::MusicPlayer - Not valid sound plugin\n");
	}
#endif

	songend = 0;
	current_plugin = NULL;
	vol = 8;
	subtune = 0;
	total_samples = 0;
	blend = 0;
	heavy = 0;
	playing = false;
	current_freq = 22050;
	songend_samples = 99999999;
	InitSound(current_freq);
#ifndef _WIN32
	mkdir(".tmpsongs", 0x1FF);
#endif
}

const char *BaseName(const char *fname)
{
	const char *ptr = &fname[strlen(fname)-1];
	while(ptr > fname && *ptr != ':' && *ptr != '\\' && *ptr != '/')
		ptr--;
	if(ptr != fname)
		ptr++;
	return ptr;
}

void hw_set_cpu(int speed);
static char *module = NULL;
static char tmpmodfile[128] = "";
extern int current_clockfreq;
extern char new_song; // Used in PlayerApp.cpp, a bit of a hack..

bool MusicPlayer::PlaySong(const char *name, char *data, int size)
{
    new_song = 5; // Used in PlayerApp.cpp, value determines how long text stays before scrolling
	memset(temp_name, 0, 64);
    strcpy(temp_name, BaseName(name));
    char* temp_ext = strrchr(temp_name,'.');
    if (temp_ext) *temp_ext = '\0';
    
    for(unsigned int i=0; i<sound_plugins.size(); i++)
	{
		if(sound_plugins[i]->can_handle(name))
		{
			int rc = -1;

			playing = false;

			cprintf("MusicPlayer::PlaySong - plugin %d\n",i);
			go_on = 0;
			SDL_LockAudio();
			if(current_plugin && current_plugin->close)
				current_plugin->close();
			audioFifo->Reset();

			if(module)
				free(module);
			module = NULL;

			if(strlen(tmpmodfile))
				remove(tmpmodfile);
			*tmpmodfile = 0;

			current_plugin = sound_plugins[i];

			if(data && current_plugin->init_data)
			{
				module = (char *)malloc(size);
				memcpy(module, data, size);
				rc = current_plugin->init_data((char *)name, module, size);
			}
			else
			if(!data && current_plugin->init_file)
			{
				rc = current_plugin->init_file((char *)name);
			}
			else
			if(!data && current_plugin->init_data)
			{
				struct stat ss;
				memset(&ss, 0, sizeof(struct stat));
				if(stat(name, &ss) == 0)
				{
					char *tmp = (char *)malloc(ss.st_size);
					FILE *fp = fopen(name, "rb");
					if(fp)
					{
						fread(tmp, 1, ss.st_size, fp);
						fclose(fp);
						rc = current_plugin->init_data((char *)name, tmp, ss.st_size);
						free(tmp);
					}
				}
			}
			else
			if(data && current_plugin->init_file)
			{
#ifndef _WIN32
				sprintf(tmpmodfile, ".tmpsongs/%s", BaseName(name));
#else
				sprintf(tmpmodfile, ".oldplay_song_%s", BaseName(name));
#endif
				FILE *fp = fopen(tmpmodfile, "wb");
				if(fp)
				{
					fwrite(data, 1, size, fp);
					fclose(fp);
					rc = current_plugin->init_file(tmpmodfile);
				}
			}


			fprintf(stderr, "MusicPlayer::PlaySong - rc = %d\n", rc);

			if(rc >= 0)
			{
				total_samples = 0;
				if(current_plugin->length > 0)
					songend_samples = (current_plugin->length * 4) / 1000 * current_plugin->freq;
				else
				if(current_plugin->length == -1)
					songend_samples = 60*6 * 4 * current_plugin->freq;
				else
					songend_samples = 0x7FFFFFFE;

				subtune = current_plugin->tune;
				audioFifo->Reset();

#ifdef GP2X
				//if(current_clockfreq > 0)
				{
					if((current_plugin->freq != current_freq) || (current_plugin->clockfreq != current_clockfreq))
					{
						SDL_UnlockAudio();
						SDL_CloseAudio();
						aopen = false;

						if(current_clockfreq < 200)
						{
							if(current_plugin->clockfreq >= 200)
								current_clockfreq = current_plugin->clockfreq;
							else
								current_clockfreq = 200;
							printf("MusicPlayer::PlaySong - clockfreq to %d\n",current_clockfreq);
							hw_set_cpu(current_clockfreq);
						}

						current_freq = InitSound(current_plugin->freq);

						if(current_plugin->clockfreq && (current_plugin->clockfreq != current_clockfreq))
						{
							current_clockfreq = current_plugin->clockfreq;
							printf("MusicPlayer::PlaySong - clockfreq to %d\n",current_clockfreq);
							hw_set_cpu(current_clockfreq);
						}

						fprintf(stderr, "MusicPlayer::PlaySong - Plugin wanted %d, gets %d\n", current_plugin->freq, current_freq);
						current_plugin->freq = current_freq;
						SDL_LockAudio();
					}
				}
				//else
#else
				if(current_plugin->freq != current_freq)
				{
					SDL_UnlockAudio();
					current_freq = InitSound(current_plugin->freq);
					fprintf(stderr, "MusicPlayer::PlaySong - Plugin wanted %d, gets %d\n", current_plugin->freq, current_freq);
					current_plugin->freq = current_freq;
					SDL_LockAudio();
				}
#endif			
				fill_audio(audioFifo, 0);
				if(!current_plugin)
					rc = -1;
				else
				{
					heavy = current_clockfreq > 200 ? 0 : (strcmp(current_plugin->plugname, "uade") == 0);
					if(!strlen(current_plugin->name))
						current_plugin->name = temp_name;
					playing = true;
				}

			}
            else {
                songend_samples = -1;
            }
			go_on = 1;
            SDL_UnlockAudio();
			return (rc >= 0);
		}
	}
	return false;
}

void MusicPlayer::SetVolume(int v)
{ 
	if(v < 0)
		v = 0;
	if(v > 10)
		v = 10;
	vol = v;

#ifdef USE_DEV_MIXER
	int vol0 = 0;
	int mvol =(((vol*0x50)/10)<<8)|((vol*0x50)/10);   
	ioctl(mixer_fd, SOUND_MIXER_WRITE_PCM, &mvol); 
#endif
}

void MusicPlayer::SetTune(int n)
{
	if(!current_plugin || !current_plugin->set_position)
		return;
    if (current_plugin->subtunes <= 1)
    {
        if (!stricmp(current_plugin->plugname,"uade"))
            return; // Hack'ish, but uade is a bitch
        // Seek n*10 seconds (n = +1,-1)
        SDL_LockAudio();
        int sec = current_plugin->set_position(n,0);
        SDL_UnlockAudio();
        int samp = total_samples + sec*current_plugin->freq*4; // I think this is right
        total_samples = samp > 0 ? samp : 0;
    }
    else
    {
        printf("MusicPlayer::SetTune - Changing subtune %d\n",n);
        new_song = 5; // See explanation in PlaySong above
        if(n < 0)
            n = 0;
        if(n >= current_plugin->subtunes)
            n = current_plugin->subtunes-1;
    
        if(subtune != n)
        {
            total_samples = 0;
            subtune = n;
            go_on = 0;
            SDL_LockAudio();
            current_plugin->set_position(0, n);
            if(!strlen(current_plugin->name))
                current_plugin->name = temp_name;
            if(current_plugin->length > 0)
                songend_samples = (current_plugin->length * 4) / 1000 * current_plugin->freq;
            else if(current_plugin->length == -1)
                songend_samples = 60*6 * 4 * current_plugin->freq;
            else
                songend_samples = 0x7FFFFFFE;
    
            audioFifo->Reset();
            fill_audio(audioFifo, 0);
            go_on = 1;
            SDL_UnlockAudio();
        }
    }
}

